import re

MARKDOWN = {
    'h1': ('\n# ', '\n'),
    'h2': ('\n## ', '\n'),
    'h3': ('\n### ', '\n'),
    'h4': ('\n#### ', '\n'),
    'h5': ('\n##### ', '\n'),
    'h6': ('\n###### ', '\n'),
    'code': ('`', '`'),
    'ul': ('', ''),
    'ol': ('', ''),
    'li': ('- ', ''),
    'blockquote': ('\n> ', '\n'),
    'em': ('**', '**'),
    'block_code': ('\n```\n', '\n```\n'),
    'span': ('', ''),
    'p': ('\n', '\n'),
    'p_with_out_class': ('\n', '\n'),
    'inline_p': ('', ''),
    'inline_p_with_out_class': ('', ''),
    'b': ('**', '**'),
    'i': ('*', '*')
}

BlOCK_ELEMENTS = (
    ('h1', '<h1.*?>(.*?)</h1>'),
    ('h2', '<h2.*?>(.*?)</h2>'),
    ('h3', '<h3.*?>(.*?)</h3>'),
    ('h4', '<h4.*?>(.*?)</h4>'),
    ('h5', '<h5.*?>(.*?)</h5>'),
    ('h6', '<h6.*?>(.*?)</h6>'),
    ('hr', '<hr/>'),
    ('blockquote', '<blockquote.*?>(.*?)</blockquote>'),
    ('ul', '<ul.*?>(.*?)</ul>'),
    ('ol', '<ol.*?>(.*?)</ol>'),
    ('block_code', '<pre.*?><code.*?>(.*?)</code></pre>'),
    ('p', '<p\s.*?>(.*?)</p>'),
    ('p_with_out_class', '<p>(.*?)</p>'),
    ('b', '<b>(.*?)</b>'),
    ('i', '<i>(.*?)</i>'))

INLINE_ELEMENTS = {
    'b': '<b>(.*?)</b>',
    'i': '<i>(.*?)</i>',
    'inline_p': '<p\s.*?>(.*?)</p>',
    'inline_p_with_out_class': '<p>(.*?)</p>',
    'code': '<code.*?>(.*?)</code>',
    'span': '<span.*?>(.*?)</span>',
    'ul': '<ul.*?>(.*?)</ul>',
    'ol': '<ol.*?>(.*?)</ol>',
    'li': '<li.*?>(.*?)</li>',
    'img': '<img.*?src="(.*?)".*?>(.*?)</img>',
    'a': '<a.*?href="(.*?)".*?>(.*?)</a>',
    'em': '<em.*?>(.*?)</em>'
}

DELETE_ELEMENTS = ['<span.*?>', '</span>', '<div.*?>', '</div>']


class Element:
    def __init__(self, start_pos, end_pos, content, tag, is_block=False):
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.content = content
        self._elements = []
        self.is_block = is_block
        self.tag = tag
        self._result = None

        if self.is_block:
            self.parse_inline()

    def __str__(self):
        wrapper = MARKDOWN.get(self.tag)
        self._result = '{}{}{}'.format(wrapper[0], self.content, wrapper[1])
        return self._result

    def parse_inline(self):
        for tag, pattern in INLINE_ELEMENTS.items():

            if tag == 'a':
                self.content = re.sub(pattern, '[\g<2>](\g<1>)', self.content)
            elif tag == 'img':
                self.content = re.sub(pattern, '![\g<2>](\g<1>)', self.content)
            elif self.tag == 'ul' and tag == 'li':
                self.content = re.sub(pattern, '- \g<1>', self.content)
            elif self.tag == 'ol' and tag == 'li':
                self.content = re.sub(pattern, '1. \g<1>', self.content)
            else:
                wrapper = MARKDOWN.get(tag)
                self.content = re.sub(pattern, '{}\g<1>{}'.format(wrapper[0], wrapper[1]), self.content)


class Tomd:
    def __init__(self, html):
        self.html = html
        self._elements = []
        self._markdown = None
        self.parse_block()
        for index, element in enumerate(DELETE_ELEMENTS):
            self._markdown = re.sub(element, '', self._markdown)

    def parse_block(self):
        for tag, pattern in BlOCK_ELEMENTS:
            for m in re.finditer(pattern, self.html, re.I | re.S | re.M):
                element = Element(start_pos=m.start(),
                                  end_pos=m.end(),
                                  content=''.join(m.groups()),
                                  tag=tag,
                                  is_block=True)
                can_append = True
                for e in self._elements:
                    if e.start_pos < m.start() and e.end_pos > m.end():
                        can_append = False
                if can_append:
                    self._elements.append(element)

        self._elements.sort(key=lambda element: element.start_pos)
        self._markdown = ''.join([str(e) for e in self._elements])

    @property
    def markdown(self):
        return self._markdown
